#!/usr/bin/env python

# Copyright (C) 2013 Jason Piper - j.piper@warwick.ac.uk
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import multiprocessing as mp
import argparse, os
from clint.textui import progress, puts_err
import pyDNase
from pyDNase import footprinting

__version__ = "0.2.0"

parser = argparse.ArgumentParser(description='Footprint the DHSs in a DNase-seq or ATAC-seq experiment using the Wellington Algorithm.')
parser.add_argument("-b","--bonferroni",action="store_true", help="Performs a bonferroni correction (default: False)",default=False)
parser.add_argument("-sh", "--shoulder-sizes", help="Range of shoulder sizes to try in format \"from,to,step\" (default: 35,36,1)",default="35,36,1",type=str)
parser.add_argument("-fp", "--footprint-sizes", help="Range of footprint sizes to try in format \"from,to,step\" (default: 11,26,2)",default="11,26,2",type=str)
parser.add_argument("-d", "--one_dimension",action="store_true", help="Use Wellington 1D instead of Wellington (default: False)",default=False)
parser.add_argument("-fdr","--FDR_cutoff", help="Write footprints using the FDR selection method at a specific FDR (default: 0.01)",default=0.01,type=float)
parser.add_argument("-fdriter", "--FDR_iterations", help="How many randomisations to use when performing FDR calculations (default: 100)",default=100,type=int)
parser.add_argument("-fdrlimit", "--FDR_limit", help="Minimum p-value to be considered significant for FDR calculation (default: -20)",default=-20,type=int)
parser.add_argument("-pv","--pv_cutoffs", help="Select footprints using a range of pvalue cutoffs (default: -10,-20,-30,-40,-50,-75,-100,-300,-500,-700",default="-10,-20,-30,-40,-50,-75,-100,-300,-500,-700",type=str)
parser.add_argument("-dm","--dont-merge-footprints",action="store_true", help="Disables merging of overlapping footprints (Default: False)",default=False)
parser.add_argument("-o","--output_prefix", help="The prefix for results files (default: <reads.regions>)",default="",type=str)
parser.add_argument("-p", help="Number of processes to use (default: uses all CPUs)",default=0,type=int)
parser.add_argument("-A",action="store_true", help="ATAC-seq mode (default: False)",default=False)
parser.add_argument("regions", help="BED file of the regions you want to footprint")
parser.add_argument("reads", help="The BAM file containing the DNase-seq reads")
parser.add_argument("outputdir", help="A writeable directory to write the results to")
clargs = parser.parse_args()

def percentile(N, percent):
    """
    Find the percentile of a list of values.

    @parameter N - is a list of values.
    @parameter percent - a float value from 0.0 to 1.0.

    @return - the percentile of the values as a float
    """
    if not N:
        return None
    N = sorted(N)
    k = (len(N)-1) * percent
    f = math.floor(k)
    c = math.ceil(k)
    if f == c:
        return float(N[int(k)])
    d0 = N[int(f)] * (c-k)
    d1 = N[int(c)] * (k-f)
    return float(d0+d1)

#Sanity check parameters from the user

def xrange_from_string(range_string):
    try:
        range_string = map(int,range_string.split(","))
        range_string = range(range_string[0],range_string[1],range_string[2])
        assert len(range_string) > 0
        return range_string
    except:
        raise ValueError

try:
    clargs.shoulder_sizes = xrange_from_string(clargs.shoulder_sizes)
    clargs.footprint_sizes = xrange_from_string(clargs.footprint_sizes)
except ValueError:
    raise RuntimeError("shoulder and footprint sizes must be supplied as from,to,step")

try:
    clargs.pv_cutoffs = map(int,clargs.pv_cutoffs.split(","))
except:
    raise RuntimeError("p-value cutoffs must be supplied as a string of numbers separated by commas")

assert 0 < clargs.FDR_cutoff < 1, "FDR must be between 0 and 1"
assert clargs.FDR_limit < 0, "FDR limit must be less than 0"
assert len([f for f in os.listdir(clargs.outputdir) if f[0] != "."]) == 0, "output directory {0} is not empty!".format(clargs.outputdir)

if not clargs.output_prefix:
    clargs.output_prefix = str(os.path.basename(clargs.reads)) + "." + str(os.path.basename(clargs.regions))

#Load reads and regions
regions = pyDNase.GenomicIntervalSet(clargs.regions)
reads = pyDNase.BAMHandler(clargs.reads,caching=False,ATAC=clargs.A)

#Create a directory for p-values and WIG output. This /should/ be OS independent
os.makedirs(os.path.join(clargs.outputdir,"p value cutoffs"))
wigout = open(os.path.relpath(clargs.outputdir) + "/" + clargs.output_prefix + ".WellingtonFootprints.wig","w")
fdrout = open(os.path.relpath(clargs.outputdir) + "/" + clargs.output_prefix + ".WellingtonFootprints.FDR.{0}.bed".format(clargs.FDR_cutoff),"w")

#Required for UCSC upload
print >> wigout, "track type=wiggle_0"

#Iterate in chromosome, basepair order
orderedbychr = [item for sublist in sorted(regions.intervals.values()) for item in sorted(sublist, key=lambda peak: peak.startbp)]
puts_err("Calculating footprints...")

if clargs.p:
    CPUs = clargs.p
else:
    CPUs = mp.cpu_count()
max_regions_cached_in_memory = 10 * CPUs
p = mp.Pool(CPUs)


def writetodisk(fp):
    #Raw WIG scores
    print >> wigout, "fixedStep\tchrom=" + str(fp.interval.chromosome) + "\t start="+ str(fp.interval.startbp) +"\tstep=1"
    print >> wigout , '\n'.join(map(str, fp.scores))
    #FDR cutoff footprints
    fdr = fp.FDR_value
    if fdr < clargs.FDR_limit:
         for footprint in fp.footprints(withCutoff=fdr,merge=clargs.dont_merge_footprints):
             print >> fdrout, footprint
    #p-value cutoff footprints
    for fpscore in clargs.pv_cutoffs:
        ofile = open(os.path.relpath(os.path.join(clargs.outputdir,"p value cutoffs")) + "/" + clargs.output_prefix + ".WellingtonFootprints.{0}.bed".format(fpscore),"a")
        for footprint in fp.footprints(withCutoff=fpscore):
            print >> ofile, footprint
        ofile.close()

def multiWellington(regions,reads,**kwargs):
    p = mp.Pool(CPUs)
    for i in progress.bar(regions):
        if clargs.one_dimension:
            fp = footprinting.wellington1D(i,reads,**kwargs)
        else:
            fp = footprinting.wellington(i,reads,**kwargs)
        p.apply_async(fp,callback = writetodisk)
        #Hold here while the queue is bigger than the number of reads we're happy to store in memory
        while p._taskqueue.qsize() > max_regions_cached_in_memory:
            pass
    p.close()
    puts_err("Waiting for the last {0} jobs to finish...".format(max_regions_cached_in_memory))
    p.join()

#TODO: Use **args or something similar to pass arguments?
multiWellington(orderedbychr,reads, shoulder_sizes = clargs.shoulder_sizes ,footprint_sizes = clargs.footprint_sizes, FDR_cutoff=clargs.FDR_cutoff,FDR_iterations=clargs.FDR_iterations,bonferroni = clargs.bonferroni)

wigout.close()



#!/usr/bin/env python

# Copyright (C) 2015 Jason Piper - j.piper@warwick.ac.uk
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import pyDNase
from clint.textui import progress, puts

parser = argparse.ArgumentParser(description='Annotates a BED file with with the number of DNase cuts in it')
parser.add_argument("-A",action="store_true", help="ATAC-seq mode (default: False)",default=False)
parser.add_argument("regions", help="BED file")
parser.add_argument("reads", help="The BAM file containing the DNase-seq data")
parser.add_argument("output", help="filename to write the output to")
args  = parser.parse_args()

reads  = pyDNase.BAMHandler(args.reads, caching=0, ATAC=args.A)
regions = pyDNase.GenomicIntervalSet(args.regions)
ofile = open(args.output,"w")

puts("Writing output...")
for i in progress.bar(regions):
    i.score   = sum([sum(j) for j in reads[i].values()])
    print >> ofile, i



#!/usr/bin/env python

# Copyright (C) 2015 Jason Piper - j.piper@warwick.ac.uk
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pyDNase, pyDNase.footprinting
import numpy as np
from clint.textui import progress
import multiprocessing as mp
import argparse

__version__ = "0.1.0"

def write_treat_to_disk(item):
    if item.results:
        for i in item.results:
            print >> treatment_output, i


def write_control_to_disk(item):
    if item.results:
        for i in item.results:
            print >> control_output, i


def xrange_from_string(range_string):
    try:
        range_string = map(int, range_string.split(","))
        range_string = range(range_string[0], range_string[1], range_string[2])
        assert len(range_string) > 0
        return range_string
    except:
        raise ValueError


class Diffwell(pyDNase.footprinting.wellington):
    def __init__(self, reads2, min_score,  *args, **kwargs):
        super(Diffwell, self).__init__(*args, **kwargs)
        self.MIN_SCORE = min_score
        self.reads2 = reads2[self.interval]

    def footprints(self, withCutoff = -20, merge = 1):
        """
        This returns reads GenomicIntervalSet with the intervals retrieved
        below the specific cutoff applied to the selected data
        """

        ranges = []

        tempMLE, templogProb = np.copy(self.lengths), np.copy(self.scores)

        while templogProb.min() < withCutoff:
            minimapos = templogProb.argmin()
            minimafplen = tempMLE[minimapos]
            minimaphalffplen = minimafplen//2
            lbound = max(minimapos-minimaphalffplen, 0)
            rbound = min(minimapos+minimaphalffplen, len(templogProb))
            ranges.append((lbound, rbound, templogProb.min(), minimafplen))
            templogProb[max(lbound-self.shoulder_sizes[-1], 0):min(rbound+self.shoulder_sizes[-1], len(templogProb))] = 1

        return_set = []
        if ranges:
            merged_ranges = []
            while len(ranges):
                # Find best score
                sorted(ranges, key=lambda x: -x[2])
                # Take the last value
                best = ranges.pop()
                merged_ranges.append(best)
                # Check for overlapping regions and remove
                new_ranges = []
                for c, d, e, f in ranges:
                    if not c <= best[1] <= d:
                        new_ranges.append([c, d, e, f])
                ranges = new_ranges
            else:
                 merged_ranges = ranges
            # Creates reads GenomicIntervalSet and adds the footprints to them
            for i in merged_ranges:
                return_set.append(((i[0] + i[1])/2, i[3]))
        return return_set

    def findDiffFP(self):
        cuts = self.reads
        forwardArray, backwardArray = cuts["+"], cuts["-"]

        cuts2 = self.reads2
        forwardArray2, backwardArray2 = cuts2["+"], cuts2["-"]

        # Adjust the FDR threshold to a minimum of withCutoff
        threshold = min(self.FDR_value, self.MIN_SCORE)

        # Find the footprints at this threshold
        offsets = self.footprints(threshold)

        # Work out the bootstrap scores for these footprints using the other data set
        best_probabilities, best_footprintsizes = pyDNase.footprinting.WellingtonC.diff_calculate(forwardArray,
                                                                                                  backwardArray,
                                                                                                  forwardArray2,
                                                                                                  backwardArray2,
                                                                                                  [i[1] for i in offsets],
                                                                                                  [i[0] for i in offsets],
                                                                                                  threshold)

        result_intervals = []

        for i in offsets:
            middle = self.interval.startbp + i[0]
            fp_halfsize = (best_footprintsizes[i[0]] // 2)
            left = middle - fp_halfsize
            right = middle + fp_halfsize
            ml_score = best_probabilities[i[0]]
            result = pyDNase.GenomicInterval(self.interval.chromosome, left, right, score=ml_score)
            result_intervals.append(result)
        return result_intervals

    def __call__(self):
        results = None
        # this is where the first round of footprinting is actually called, as self.scores invoked the footprinting
        if min(self.scores) < self.MIN_SCORE:
            if min(self.scores) < self.FDR_value:
                results = self.findDiffFP()
        self.results = results
        return self



parser = argparse.ArgumentParser(description='Scores Differential Footprints using Wellington-Bootstrap.')
parser.add_argument("-fp", "--footprint-sizes",
                    help="Range of footprint sizes to try in format \"from,to,step\" (default: 11,26,2)",
                    default="11,26,2",
                    type=str)
parser.add_argument("-fdr","--FDR_cutoff",
                    help="Detect footprints using the FDR selection method at a specific FDR (default: 0.01)",
                    default=0.01,
                    type=float)
parser.add_argument("-fdriter", "--FDR_iterations",
                    help="How many randomisations to use when performing FDR calculations (default: 100)",
                    default=100,
                    type=int)
parser.add_argument("-fdrlimit", "--FDR_limit",
                    help="Minimum p-value to be considered significant for FDR calculation (default: -20)",
                    default=-20,
                    type=int)
parser.add_argument("-p", "--processes", help="Number of processes to use (default: uses all CPUs)",
                    default=0,
                    type=int)
parser.add_argument("-A", action="store_true", help="ATAC-seq mode (default: False)", default=False)
parser.add_argument("treatment_bam", help="BAM file for treatment")
parser.add_argument("control_bam", help="BAM file for control")
parser.add_argument("bedsites", help="BED file of genomic locations to search in")
parser.add_argument("treatment_only_output", help="File to write treatment specific fooprints scores to")
parser.add_argument("control_only_output", help="File to write control specific footprint scores to")
args = parser.parse_args()

# Sanity check parameters from the user

try:
    args.footprint_sizes = xrange_from_string(args.footprint_sizes)
except ValueError:
    raise RuntimeError("Footprint sizes must be supplied as from,to,step")

assert 0 < args.FDR_cutoff < 1, "FDR must be between 0 and 1"
assert args.FDR_limit < 0, "FDR limit must be less than 0"

# Treatment
reads2 = pyDNase.BAMHandler(args.treatment_bam, caching=0, ATAC=clargs.A)
# Control
reads1 = pyDNase.BAMHandler(args.control_bam, caching=0, ATAC=clargs.A)
# Regions of Interest
regions = pyDNase.GenomicIntervalSet(args.bedsites)
# Output
treatment_output = open(args.treatment_only_output, "w", buffering=1)
control_output   = open(args.control_only_output, "w", buffering=1)

# Determine Number of CPUs to use
if args.processes:
    CPUs = args.processes
else:
    CPUs = mp.cpu_count()
# NOTE: This roughly scales at about 450mb per 300 regions held in memory
max_regions_cached_in_memory = 50 * CPUs
p = mp.Pool(CPUs)

print "Performing differential footprinting..."

for i in progress.bar(regions):
    # Make sure the interval is actually big enough to footprint to begin with
    if len(i) < 120:
        i.startbp -= 60
        i.endbp += 60
    # Make the optional arguments
    fp_args = {'footprint_sizes': args.footprint_sizes, 'FDR_cutoff': args.FDR_cutoff, 'FDR_iterations': args.FDR_iterations}
    # Perform both comparisons - A against B and B against A
    fp = Diffwell(reads2=reads1, min_score=args.FDR_limit, interval=i, reads=reads2, **fp_args)
    fp2 = Diffwell(reads2=reads2, min_score=args.FDR_limit, interval=i, reads=reads1, **fp_args)
    # Push these tasks to the queue
    p.apply_async(fp, callback=write_treat_to_disk)
    p.apply_async(fp2, callback=write_control_to_disk)
    # Hold here while the queue is bigger than the number of reads we're happy to store in memory
    while p._taskqueue.qsize() > max_regions_cached_in_memory:
        pass



#!/usr/bin/env python

# Copyright (C) 2015 Jason Piper - j.piper@warwick.ac.uk
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import pyDNase
import math
from clint.textui import progress, puts

parser = argparse.ArgumentParser(description='Annotates a set of DHSs with the dDHS score (He et al. 2012)')
parser.add_argument("-l", action="store_true", help="low RAM mode (disables caching) (Default: False)",default=False)
parser.add_argument("-A",action="store_true", help="ATAC-seq mode (default: False)",default=False)
parser.add_argument("regions", help="The set of BED files you wish to annotate with dDHS scores")
parser.add_argument("treat_dhs", help="The DHSs belonging to the Treatment")
parser.add_argument("control_dhs", help="The DHSs belonging to the control")
parser.add_argument("reads_treat", help="The BAM file containing the Treatment DNase-seq data")
parser.add_argument("reads_control", help="The BAM file containing the Control DNase-seq data")
parser.add_argument("output", help="filename to write the output to")
args  = parser.parse_args()

reads_treat   = pyDNase.BAMHandler(args.reads_treat, caching = not args.l, ATAC=args.A)
reads_control = pyDNase.BAMHandler(args.reads_control, caching = not args.l, ATAC=args.A)
treat_dhs     = pyDNase.GenomicIntervalSet(args.treat_dhs)
control_dhs   = pyDNase.GenomicIntervalSet(args.control_dhs)
regions       = pyDNase.GenomicIntervalSet(args.regions)

treat_total_cuts   = 0
control_total_cuts = 0
treat_base_pairs   = 0
control_base_pairs = 0

puts("Calculating enrichment for Treatment")
for i in progress.bar(treat_dhs):
    treat_total_cuts += sum([sum(j) for j in reads_treat[i].values()])
    treat_base_pairs += len(i)

puts("Calculating enrichment for Control")
for i in progress.bar(control_dhs):
    control_total_cuts += sum([sum(j) for j in reads_control[i].values()])
    control_base_pairs += len(i)

treat_base_pairs   = float(treat_base_pairs)
control_base_pairs = float(control_base_pairs)
treat_total_cuts   = float(treat_total_cuts)
control_total_cuts = float(control_total_cuts)

ofile = open(args.output,"w")

puts("Calculating dDHS scores...")
for i in progress.bar(regions):
    treat_cuts   =  sum([sum(j) for j in reads_treat[i].values()])
    control_cuts =  sum([sum(j) for j in reads_control[i].values()])
    LHS = math.sqrt(  treat_cuts /(  treat_total_cuts / treat_base_pairs))
    RHS = math.sqrt(control_cuts /(control_total_cuts / control_base_pairs))
    i.score = LHS - RHS
    print >> ofile, i
ofile.close()



cimport cython
from libc.math cimport ceil
from libc.stdlib cimport malloc, free
import random
random.seed("Congratulations on reading the source code - you win a prize!")

cdef extern from "WellingtonC.h":
    struct tuple2:
         float * fpscores
         unsigned int * mles
    double bdtrc(int, int, float)
    double bdtr(int, int, float)
    tuple2 * wellington(unsigned int *, unsigned int *, unsigned int,unsigned int *, unsigned int,unsigned int *, unsigned int)
    tuple2 * wellington1D(unsigned int *, unsigned int *, unsigned int,unsigned int *, unsigned int,unsigned int *, unsigned int)
    tuple2 * diff_wellington(unsigned int * f,  unsigned int * r, unsigned int * f2,  unsigned int * r2, unsigned int length, unsigned int * offsets, unsigned int * widths, unsigned int num_offsets,float threshold)

cdef float logsf(int a,int b,float c):
    return bdtrc(a, b, c)
cdef float logcdf(int a,int b,float c):
    return bdtr(a,b,c)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
cpdef float percentile(list N, float percent):
    """
    Find the percentile of a list of values.

    @parameter N - is a list of values.
    @parameter percent - a float value from 0.0 to 1.0.

    @return - the percentile of the values as a float
    """
    cdef float k,d0,d1
    cdef unsigned int f, c
    if not N:
        return None
    N = sorted(N)
    k = (len(N)-1) * percent
    f = <int>k
    c = <int>ceil(k)
    if f == c:
        return float(N[int(k)])
    d0 = N[f] * (c-k)
    d1 = N[c] * (k-f)
    return d0+d1

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def diff_calculate(list forwardArray, list backwardArray,list forwardArray2, backwardArray2, footprint_sizes, offset_positions,float threshold):
    cdef unsigned int asize = len(forwardArray)
    cdef unsigned int *farr  = <unsigned int *> malloc(asize * sizeof(unsigned int))
    cdef unsigned int *rarr  = <unsigned int *> malloc(asize * sizeof(unsigned int))
    cdef unsigned int *farr2  = <unsigned int *> malloc(asize * sizeof(unsigned int))
    cdef unsigned int *rarr2  = <unsigned int *> malloc(asize * sizeof(unsigned int))
    cdef unsigned int *fpsize  = <unsigned int *> malloc(len(footprint_sizes) * sizeof(unsigned int))
    cdef unsigned int *offsets  = <unsigned int *> malloc(len(offset_positions) * sizeof(unsigned int))
    for k in range(asize):
        farr[k] = forwardArray[k]
        rarr[k] = backwardArray[k]
        farr2[k] = forwardArray2[k]
        rarr2[k] = backwardArray2[k]

    for index, size in enumerate(footprint_sizes):
        fpsize[index] = size
    for index, size in enumerate(offset_positions):
        offsets[index] = size
    #diff_wellington(unsigned int * f,  unsigned int * r, unsigned int * f2,  unsigned int * r2, unsigned int length, unsigned int * offsets, unsigned int * widths, unsigned int num_offsets)
   # tuple2 * test = diff_wellington(farr,  rarr, farr2,  rarr2, asize, offsets, fpsize, len(footprint_sizes))
    cdef tuple2 * test = diff_wellington(farr,  rarr, farr2,  rarr2, asize, offsets, fpsize, len(footprint_sizes),threshold)
    #Push the values from C arrays back into Python lists
    cdef list m,f
    m, f  = [], []
    cdef unsigned int j
    for j in range(asize):
        f.append(test.fpscores[j])
        m.append(int(test.mles[j]))

    #Clean everything up
    free(test.mles)
    free(test.fpscores)
    free(test)
    free(farr)
    free(rarr)
    free(farr2)
    free(rarr2)
    free(fpsize)
    free(offsets)

    #Return the Scores and the FP parameters
    return(f,m)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def calculate(FDR, list forwardArray_in, list backwardArray_in, footprint_sizes, shoulder_sizes, bonferroni_factor):

    #If we're looking for an FDR, shuffle the data here.

    if FDR:
        forwardArray, backwardArray = forwardArray_in[:], backwardArray_in[:]
        random.shuffle(forwardArray)
        random.shuffle(backwardArray)
    else:
        forwardArray, backwardArray = forwardArray_in, backwardArray_in

    cdef unsigned int asize = len(forwardArray)

    #Copy the forward and reverse reads and the shoulder and fp sizes into the C arrays
    cdef unsigned int * farr  = <unsigned int *> malloc(asize * sizeof(unsigned int))
    cdef unsigned int * rarr  = <unsigned int *> malloc(asize * sizeof(unsigned int))
    cdef unsigned int * shsize  = <unsigned int *> malloc(len(shoulder_sizes) * sizeof(unsigned int))
    cdef unsigned int * fpsize  = <unsigned int *> malloc(len(footprint_sizes) * sizeof(unsigned int))
    for k in range(asize):
        farr[k] = forwardArray[k]
        rarr[k] = backwardArray[k]

    for index, size in enumerate(shoulder_sizes):
        shsize[index] = size
    for index, size in enumerate(footprint_sizes):
        fpsize[index] = size

    #Farm off the Computation to C
    cdef tuple2 * test = wellington(farr,rarr,asize,shsize,len(shoulder_sizes),fpsize,len(footprint_sizes))

    #Push the values from C arrays back into Python lists
    cdef list m,f
    m, f  = [], []
    cdef unsigned int j
    cdef int minlimit = -1000
    for j in range(asize):
        if bonferroni_factor:
            f.append(min(0,test.fpscores[j] - bonferroni_factor))
        else:
            f.append(max(test.fpscores[j],minlimit))
        m.append(int(test.mles[j]))

    #Clean everything up
    free(test.mles)
    free(test.fpscores)
    free(test)
    free(farr)
    free(rarr)
    free(shsize)
    free(fpsize)

    #Return the Scores and the FP parameters
    return(f,m)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nonecheck(False)
def calculate1D(FDR, list forwardArray_in, list backwardArray_in, footprint_sizes, shoulder_sizes, bonferroni_factor):

    #If we're looking for an FDR, shuffle the data here.

    if FDR:
        forwardArray, backwardArray = forwardArray_in[:], backwardArray_in[:]
        random.shuffle(forwardArray)
        random.shuffle(backwardArray)
    else:
        forwardArray, backwardArray = forwardArray_in, backwardArray_in

    cdef unsigned int asize = len(forwardArray)

    #Copy the forward and reverse reads and the shoulder and fp sizes into the C arrays
    cdef unsigned int * farr  = <unsigned int *> malloc(asize * sizeof(unsigned int))
    cdef unsigned int * rarr  = <unsigned int *> malloc(asize * sizeof(unsigned int))
    cdef unsigned int * shsize  = <unsigned int *> malloc(len(shoulder_sizes) * sizeof(unsigned int))
    cdef unsigned int * fpsize  = <unsigned int *> malloc(len(footprint_sizes) * sizeof(unsigned int))
    for k in range(asize):
        farr[k] = forwardArray[k]
        rarr[k] = backwardArray[k]

    for index, size in enumerate(shoulder_sizes):
        shsize[index] = size
    for index, size in enumerate(footprint_sizes):
        fpsize[index] = size

    #Farm off the Computation to C
    cdef tuple2 * test = wellington1D(farr,rarr,asize,shsize,len(shoulder_sizes),fpsize,len(footprint_sizes))

    #Push the values from C arrays back into Python lists
    cdef list m,f
    m, f  = [], []
    cdef unsigned int j
    cdef int minlimit = -1000
    for j in range(asize):
        if bonferroni_factor:
            f.append(min(0,test.fpscores[j] - bonferroni_factor))
        else:
            f.append(max(test.fpscores[j],minlimit))
        m.append(int(test.mles[j]))

    #Clean everything up
    free(test.mles)
    free(test.fpscores)
    free(test)
    free(farr)
    free(rarr)
    free(shsize)
    free(fpsize)

    #Return the Scores and the FP parameters
    return(f,m)



#!/usr/bin/env python

# Copyright (C) 2013 Jason Piper - j.piper@warwick.ac.uk
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import matplotlib.pyplot as plt
import pyDNase
from pyDNase.footprinting import wellington

#Load test data
reads = pyDNase.BAMHandler(pyDNase.example_reads())
regions = pyDNase.GenomicIntervalSet(pyDNase.example_regions())

#Plot cuts data
plt.plot(reads[regions[0]]["+"],c="red")
plt.plot([-i for i in reads[regions[0]]["-"]],c="blue")

#Footprint and plot the results
footprinter = wellington(regions[0],reads)
plt.plot(footprinter.scores,c="black")

plt.show()


# Copyright (C) 2013 Jason Piper - j.piper@warwick.ac.uk
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from . import _version
__version__ = _version.__version__

import os, math, pysam
from clint.textui import progress, puts_err
import sqlite3 as lite
import tempfile, warnings, pickle

def example_reads():
    """
    returns the path to the example BAM file
    """
    return os.path.join(os.path.join(os.path.dirname(__file__), "data"),"example.bam")

def example_regions():
    """
    returns the path to the example BED file
    """
    return os.path.join(os.path.join(os.path.dirname(__file__), "data"),"example.bed")


class BAMHandler(object):
    """
    The object that provides the interface to DNase-seq data help in a BAM file
    """
    def __init__(self, filePath, caching=True, chunkSize=1000, ATAC=False):
        """Initializes the BAMHandler with a BAM file

        Args:
            filePath (str): the path of a sorted, indexed BAM file from a DNase-seq experiment
        Kwargs:
            chunkSize (int): and int of the size of the regions to load if caching (default: 1000)
            caching (bool): enables or disables read caching (default: True)
        Raises:
            IOError
        """
        try:
            self.samfile = pysam.Samfile(filePath)
        except IOError:
            errorString = "Unable to load BAM file:{0}".format(filePath)
            raise IOError(errorString)

        #Initialise the empty DNase cut cache with the chromosome names from the BAM file
        self.purge_cache()
        #Do not change the CHUNK_SIZE after object instantiation!
        self.CHUNK_SIZE  = chunkSize
        self.CACHING     = caching
        
        if ATAC:
            self.loffset = -5
            self.roffset = +4
        else:
            self.loffset = 0
            self.roffset = 0

    def __addCutsToCache(self,chrom,start,end):
        """Loads reads from the BAM file into the cutCache. Will not check if reads are already there.

        Args:
            chrom (str): The chromosome
            start (int): The start of the interval
            end (int): The end of the interval
        """
        for alignedread in self.samfile.fetch(chrom, max(start, 0), end):
            if alignedread.is_reverse:
                a = int(alignedread.aend)
                if a <= end +1:
                    self.cutCache[chrom]["-"][a] = self.cutCache[chrom]["-"].get(a, 0) + 1
            else:
                a = int(alignedread.pos) -1
                if a >= start:
                    self.cutCache[chrom]["+"][a] = self.cutCache[chrom]["+"].get(a, 0) + 1
        self.lookupCache[chrom].append(start)

    def __lookupReadsUsingCache(self,startbp,endbp,chrom):
        """Looks up the DNase cut information from the cutCache and returns as a dictionary (private method)

        Args:
            startbp (int): The start of the interval
            endbp (int): The end of the interval
            chrom (str): The chromosome

        """
        #Expand the region to the nearest CHUNK_SIZE and load these reads if they aren't in the cache
        lbound = int(math.floor(startbp / float(self.CHUNK_SIZE)) * float(self.CHUNK_SIZE))
        ubound = int(math.ceil(endbp / float(self.CHUNK_SIZE)) * float(self.CHUNK_SIZE))
        for i in range(lbound,ubound,self.CHUNK_SIZE):
            if i not in self.lookupCache[chrom]:
                self.__addCutsToCache(chrom,i,i+self.CHUNK_SIZE)
        #Fills in with zeroes where the hash table contains no information for each strand.
        fwCutArray  = [self.cutCache[chrom]["+"].get(i, 0) for i in range(startbp + self.loffset ,endbp + self.loffset)]
        revCutArray = [self.cutCache[chrom]["-"].get(i, 0) for i in range(startbp + self.roffset, endbp + self.roffset)]
        return {"+":fwCutArray,"-":revCutArray}

    def __lookupReadsWithoutCache(self,startbp,endbp,chrom):
        """Loads reads from the BAM file directly from disk, ignoring the cache (private method)

        Args:
            startbp (int): The start of the interval
            endbp (int): The end of the interval
            chrom (str): The chromosome
        """
        tempcutf = {}
        tempcutr = {}
        for alignedread in self.samfile.fetch(chrom, max(startbp, 0), endbp):
            if alignedread.is_reverse:
                a = int(alignedread.aend)
                if a <= endbp +1:
                    tempcutr[a] = tempcutr.get(a, 0) + 1
            else:
                a = int(alignedread.pos) - 1
                if a >= startbp:
                    tempcutf[a] =tempcutf.get(a, 0) + 1
        fwCutArray  = [tempcutf.get(i, 0) for i in range(startbp + self.loffset ,endbp + self.loffset)]
        revCutArray = [tempcutr.get(i, 0) for i in range(startbp + self.roffset, endbp + self.roffset)]
        return {"+":fwCutArray,"-":revCutArray}

    def purge_cache(self):
        """
        Wipes the internal cache - useful if you need finer grain control over caching.
        """

        #Initialise the empty DNase cut cache with the chromosome names from the BAM file
        self.cutCache = {}
        #This helps us differentiate between what's been looked up and when there's just no reads
        self.lookupCache = {}
        for i in self.samfile.references:
            self.cutCache[i]    = {"+":{},"-":{}}
            self.lookupCache[i] = []

    def get_cut_values(self,vals):
        """Return a dictionary with the cut counts. Can be used in two different ways:

        You can either use a string or a GenomicInterval to query for cuts.
        Returns reads dict with "+" corresponding to the +ve strand and "-" has the data with the -ve strand (rotated 180 degrees)

        Args:
            vals: either a string with the format "chr18,500:600,+" or a GenomicInterval object

        >>> BAMHandler(example_reads())["chr6,170863142,170863150,+"]
        {'+': array([ 1,  0,  0,  0,  1, 11,  1,  0]), '-': array([0, 1, 0, 0, 1, 0, 0, 1])}
        >>> BAMHandler(example_reads())["chr6,170863142,170863150,-"]
        {'+': array([1, 0, 0, 1, 0, 0, 1, 0]), '-': array([ 0,  1, 11,  1,  0,  0,  0,  1])}
        """

        if isinstance(vals, GenomicInterval):
            chrom   = vals.chromosome
            startbp = vals.startbp
            endbp   = vals.endbp
            flip    = vals.strand

        elif isinstance(vals, str):
            try:
                chrom,startbp,endbp,flip   = vals.split(",")
                startbp = int(startbp)
                endbp   = int(endbp)
                assert(flip in ["+", "-"])
            except:
                raise ValueError("Malformed query string")

        else:
            raise TypeError("Lookup must be a string or a GenomicInterval")

        if self.CACHING:
            retval = self.__lookupReadsUsingCache(startbp,endbp,chrom)
        else:
            retval = self.__lookupReadsWithoutCache(startbp,endbp,chrom)

        if flip is "-":
            retval["+"], retval["-"] = retval["-"][::-1], retval["+"][::-1]
        return retval

    def __getitem__(self,vals):
        """
        Wrapper for get_cut_values
        """
        return self.get_cut_values(vals)

    def FOS(self,interval,bgsize=35):
        """Calculates the Footprint Occupancy Score (FOS) for a Genomicinterval. See Neph et al. 2012 (Nature) for full details.
        
        Args:
            interval (GenomicInterval): The interval that you want the FOS for

        Kwargs:
            bgsize (int): The size of the flanking region to use when calculating the FOS (default: 35)

        Returns:
            A float with the FOS - returns -1 if it can't calculate it
        """

        cuts = self["{0},{1},{2},{3}".format(interval.chromosome,interval.startbp-bgsize,interval.endbp+bgsize,interval.strand)]
        forwardArray, backwardArray     = cuts["+"], cuts["-"]
        cutArray     = [forwardArray[i] + backwardArray[i] for i in range(len(forwardArray))]

        leftReads   = float(sum(cutArray[:bgsize]))
        centreReads = float(sum(cutArray[bgsize:-bgsize]))
        rightReads  = float(sum(cutArray[-bgsize:]))

        #Here we normalise by region length
        leftReads   /= (bgsize * 1.0)
        centreReads /= (len(interval) * 1.0)
        rightReads  /= (bgsize * 1.0)

        try:
            return ( (centreReads+1.0) / (leftReads + 1.0) ) + ( (centreReads+1.0)/(rightReads + 1.0))
        except BaseException:
            #If it can't calculate the score, return a sentinel value
            return -1


class GenomicIntervalSet(object):
    """Container class which stores and allow manipulations of large numbers of GenomicInterval objects.
    Essentially a way of storing and sorting BED files.
    """

    def __init__(self,filename = None):
        """
        Inits GenomicIntervalSet. You can also specify a BED file path to load the intervals from

        Kwargs:
            filename (str): the path to a BED file to initialize the intervals with

        If no ``filename`` provided, then the set will be empty
        """
        self.intervals = {}

        if filename:
            self.loadBEDFile(filename)

    def loadBEDFile(self,filename):
        """
        Adds all the intervals in a BED file to this GenomicIntervalSet.
        We're quite naughty here and allow some non-standard BED formats (along with the official one):

        chrom chromStart chromEnd
        chrom chromStart chromEnd strand
        chrom chromStart chromEnd name score strand

        Any whitespace (tabs or spaces) will be considered separators, so spaces in names cause a problem!

        .. note::
            If you don't supply a strand, we infer that it's +ve.

        Args:
            filename: the path to a BED file to load

        Raises:
            IOError
        """
        try:
            BEDfile = open(filename, 'rU')
        except IOError:
            errorString = "Cannot load BED file: {0}".format(filename)
            raise IOError(errorString)

        puts_err("Reading BED File...")

        #This is done so that if a malformed BED record is detected, no intervals are loaded.
        records = []
        
        intervalCount = max(enumerate(open(filename)))[0] + 1
        for _ in progress.bar(range(intervalCount)):
            line    = BEDfile.readline()
            #Skip lines in the bed files which are UCSC track metadata or comments
            if not self.__isBEDHeader(line):
                records.append(self.__parseBEDString(line))

        for i in records:
            self.__addInterval(GenomicInterval(i[0], i[1], i[2], i[3], i[4], i[5]))

        BEDfile.close()

    def __malformedBEDline(self,BEDString):
        """
        Raises an exception and prints the offending BED string

        Raises:
            Exception
        """
        #TODO: Make a new exception class, something like malformedBEDException?
        exceptionString = "Malformed BED line: {0}".format(BEDString)
        raise Exception(exceptionString)

    def __isBEDHeader(self,string):
        """
        Returns True/False whether a line in a bed file should be ignored according to
        http://genome.ucsc.edu/goldenPath/help/customTrack.html#TRACK
        """
        if string[0] == "#":
            return True
            
        headers = ["name","description","type","visibility","color","itemRgb","useScore","group",
                   "priority","db","offset","maxItems","url","htmlUrl","bigDataUrl","track","browser"]
                   
        for each in headers:
            if string.startswith(each):
                return True
        return False

    def __parseBEDString(self,BEDString):
        """
        Parses the following BED formats
        We're quite naughty here and allow some non-standard BED formats (along with the official one):

        chrom chromStart chromEnd
        chrom chromStart chromEnd strand
        chrom chromStart chromEnd name score strand

        Returns:
            (chrom,startbp,endbp,label,score,strand)

        Raises:
            Exception
        """
        BEDSplit = BEDString.split()

        #Sanity check
        if len(BEDSplit) not in [3,4,6]:
            self.__malformedBEDline(BEDString)

        #Default if only Chrom Start End is detected
        try:
            chrom   = BEDSplit[0]
            startbp = int(BEDSplit[1])
            endbp   = int(BEDSplit[2])
        except:
            self.__malformedBEDline(BEDString)

        label = 0
        score = 0
        strand = "+"

        if len(BEDSplit) is 4:
            if BEDSplit[3] in ["+", "-"]:
                strand = BEDSplit[3]
            else:
                self.__malformedBEDline(BEDString)

        if len(BEDSplit) is 6:
            label  = BEDSplit[3]
            try:
                score = float(BEDSplit[4])
            except ValueError:
                self.__malformedBEDline(BEDString)
            if BEDSplit[5] in ["+", "-"]:
                strand = BEDSplit[5]
            else:
                self.__malformedBEDline(BEDString)

        return chrom,startbp,endbp,label,score,strand

    def __len__(self):
        """
        Return the number of intervals in the set
        """
        intervals = 0
        for each in self.intervals.values():
            intervals += len(each)
        return intervals

    def __iter__(self):
        """
        Iterates over the intervals in the order that the intervals were generated
        """
        for each in sorted(sum(self.intervals.values(),[]), key=lambda peak: peak.importorder):
            yield each

    def __getitem__(self, i):
        """
        Indexes the intervals in the order that the intervals were generated
        """
        return sorted(sum(self.intervals.values(),[]), key=lambda peak: peak.importorder)[i]

    def __delitem__(self,i):
        """
        Deletes an interval from the set using the position i
        """
        pos = self.intervals[self[i].chromosome].index(self[i])
        del self.intervals[self[i].chromosome][pos]

    def __iadd__(self, other):
        """
        Adds all the intervals from an other GenomicIntervalSet or GenomicInterval to this one.

        Args:
            other: either a GenomicInterval or GenomicIntervalSet to be added
        Raises:
            TypeError: A GenomicInterval or GenomicIntervalSet wasn't supplied.
        """
        if isinstance(other,GenomicIntervalSet):
            for i in other:
                self.__addInterval(i)
        elif isinstance(other,GenomicInterval):
            self.__addInterval(other)
        else:
            raise TypeError("Can only add GenomicInterval or GenomicIntervalSet objects to existing GenomicIntervalSet")
        return self

    def __addInterval(self, other):
        """
        Adds a GenomicInterval to the set

        Args:
            other (GenomicInterval): The GenomicInterval to be added.
        """
        if other.chromosome not in self.intervals:
            self.intervals[other.chromosome] = []
        self.intervals[other.chromosome].append(other)

    def resizeRegions(self,toSize):
        """
        Resized all GenomicIntervals to a specific size

        Args:
            toSize: an int of the size to resize all intervals to
        """
        if not type(toSize) == int:
            ValueError("Can only resize intervals to integers")

        for i in self:
            xamount = toSize-(i.endbp-i.startbp)//2
            i.startbp -= xamount
            i.endbp   += xamount
            if (i.endbp-i.startbp) > toSize*2:
                i.endbp -= 1
    def __str__(self):
        return ''.join(str(i) +"\n" for i in self)

class GenomicInterval(object):
    """
    Basic Object which describes reads region of the genome
    """

    #This counts how many GenomicInterval objects have been created
    counter = 0

    def __init__(self, chrom, start, stop, label = 0,score = 0,strand="+"):
        """
        Initialization routine

        Args:
            chrom (str): the chromosome
            
            start (int): the start of the interval
            
            stop (int): the end of the interval

        Kwargs:
            label: The name of the interval (will be given an automatic name if none entered)
            
            score (float): the score of the interval (default: 0)
            
            strand (str): the strand the interval is on (default: "+")
        """

        self.__class__.counter += 1
        self.importorder = self.__class__.counter

        self.chromosome = str(chrom)
        self.startbp    = int(start)
        self.endbp      = int(stop)
        self.strand     = str(strand)
        self.score      = float(score)

        if self.startbp > self.endbp:
            raise Exception("Start location of GenomicInterval is larger than end location!")

        # This is from reads bygone era where we ordered the intervals by import order
        # self.score      = self.__class__.counter

        #This makes up reads fake name if one doesn't exist in the BED file
        if label:
            self.label = str(label)
        else:
            self.label = "Unnamed{0}".format(self.__class__.counter)

        #This contains anything else you want to store about the interval
        self.metadata = {}

    def __str__(self):
        """
        BED representation of the interval
        """
        return "{0}\t{1}\t{2}\t{3}\t{4}\t{5}".format(self.chromosome, self.startbp, self.endbp, self.label, self.score, self.strand)

    def __len__(self):
        """
        Returns the length of the GenomicInterval in basepairs
        """
        return self.endbp - self.startbp

class FASTAHandler(object):
    def __init__(self, fasta_file, vcf_file = None):
        self.ffile = pysam.Fastafile(fasta_file)
        self.conn = None
        if vcf_file:
            self.conn = lite.connect(tempfile.NamedTemporaryFile().name)
            with open(vcf_file, 'r') as f:
                records = [(x[0],x[1],x[3],x[4]) for x in (x.split() for x in f if x[0] != "#")]
                with self.conn:
                    cur = self.conn.cursor()
                    cur.execute("CREATE TABLE SNPS(chr TEXT,pos INT, ref TEXT, mut TEXT)")
                    cur.executemany("INSERT INTO SNPS VALUES(?,?,?,?)",records)
                #Manually remove these, as they potentially could be large in memory
                del(records)
    def sequence(self,interval):
        sequence_string = self.ffile.fetch(interval.chromosome,interval.startbp,interval.endbp).upper()
        if not self.conn:
          #  if interval.strand != "+":
           #     sequence_string = sequence_string[::-1]
            return str(sequence_string)
        else:
            query_string = "SELECT chr, pos - ? - 1 as offset,ref,mut FROM SNPS WHERE chr=? and pos BETWEEN ? and ?"
            snps = self.conn.cursor().execute(query_string,(interval.startbp,interval.chromosome,interval.startbp,interval.endbp)).fetchall()
            sequence_list = [i for i in sequence_string]
            for i in snps:
                if sequence_list[i[1]] != i[2]:
                    warnings.warn("MISMATCH IN REF TO SNP - WHAT HAVE YOU DONE?")
                else:
                    #str needed as sqlite returns unicode
                    sequence_list[i[1]] = str(i[3])
            return "".join(sequence_list)

class BiasCalculator(object):
    def __init__(self,bias_file=None):
        if bias_file is None:
            #Load the genomic IMR90 bias from Shirley
            bias_file = open(os.path.join(os.path.join(os.path.dirname(__file__), "data"),"IMR90_6mer.pickle"))
        self.biasdict = pickle.load(bias_file)
    def bias(self,sequence):
        """
        NOTE:   Because bias is calculated from  the centre of a 6-mer,
                the data will be missing the values for the first and last 3 bases
        """
        #Split sequence into consituent 6-mers
        sequence_chunks = [sequence[i:i+6] for i in range(len(sequence)-5)]
        #Look up values of these 6-mers in the precomputed bias database
        fw_bias =  [float(self.biasdict[i]["forward"])for i in sequence_chunks]
        rv_bias =  [float(self.biasdict[i]["reverse"])for i in sequence_chunks]
        #FIXME: Pickled data should use "+" and "-" and not forward and reverse to prevent confusion here
        #FIXME: Fix the pickled data - the reverse positions are off by one!
        return {"+": fw_bias, "-":rv_bias}

class BAMHandlerWithBias(BAMHandler):
    def __init__(self, sequence_object, *args, **kwargs):
        super(BAMHandlerWithBias, self).__init__(*args, **kwargs)
        self.sequence_data = sequence_object
        self.bias_data     = BiasCalculator()

    def __getitem__(self,interval):
        if not isinstance(interval,GenomicInterval):
            raise TypeError("Sorry, but we only support GenomicInterval querying for the Bias Handler at the moment")
        #Note: This is pretty Hacky!
        interval.startbp -= 3
        interval.endbp   += 3
        #Get the sequence data
        bias_values = self.bias_data.bias(self.sequence_data.sequence(interval))
        interval.startbp += 3
        interval.endbp   -= 3
        bias_values["+"] = bias_values["+"][1:]
        bias_values["-"] = bias_values["-"][:-1]
        cuts = self.get_cut_values(interval)

        #Nomenclature used below is that in the He. et al Nature Methods Paper
        #These are N_i^s - note we are using an entire hypersensitive site, and not 50bp like the paper
        N_i = {"+":sum(cuts["+"]) ,"-":sum(cuts["-"])}

        #bias_values are y_i
        for dir in ("-","+"):
            bias_values[dir] = [float(i)/sum(bias_values[dir]) for i in bias_values[dir]]

        #Stupid pass-by-reference
        predicted_cuts = {"+":cuts["+"][:],"-":cuts["-"][:]}
        #Calculate the predicted counts (nhat_i, which is N_i * y_i) based on n-mer cutting preference
        for dir in ("-","+"):
            #For each base
            for num, val in enumerate(predicted_cuts[dir]):
                #Multiply the total number of observed cuts by the bias value
                predicted_cuts[dir][num] = bias_values[dir][num] * N_i[dir]

        #Now we normalised the observed cuts by the expected
        for dir in ("-","+"):
            #For each base
            for num, val in enumerate(predicted_cuts[dir]):
                #Divide the number of observed cuts by the bias value
                pass
                #predicted_cuts[dir][num] = (cuts[dir][num] + 1.0)  / (val + 1.0)
        if interval.strand == "-":
            # That's numberwang, let's rotate the board!
            predicted_cuts["+"], predicted_cuts["-"] = predicted_cuts["-"][::-1], predicted_cuts["+"][::-1]
        return predicted_cuts



#!/usr/bin/env python

# Copyright (C) 2015 Jason Piper - j.piper@warwick.ac.uk
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os, argparse, tempfile, operator, pickle
from collections import defaultdict
import pybedtools, pysam

def rev_comp(str):
	"""
	Given a DNA string, returns the reverse complement
	
	>>> rev_comp("AATTGGCC")
	'GGCCAATT'
	"""
	rev_dic = {'A':'T','G':'C','C':'G','T':'A'}
	return ''.join([rev_dic[i] for i in str[::-1]])

def get_kmers(file_path):
	"""
	Counts the number of occurences of each sequence in a FASTA file 
	and returns a dictionary with the format {sequence:count}, e.g. {'AATTCC':5}
	
	Note: Ignores anything with Ns in, and treats all DNA as uppercase (i.e. not repeatmasked)
	"""
	kmers = defaultdict(int)
	with open(file_path,'r') as fasta:
		for i in fasta:
			if i[0] != ">":
				if 'N' not in i:
					kmers[i.strip().upper()] +=1
	return kmers

def generate_6mer_bed(bam_file, gdict):
	"""
	Loads data from a BAM file, and prints the 6-mers from the 5' ends of the reads
	as a BED file to a temp file, returns the file path of this temp file
	"""
	outfile = tempfile.NamedTemporaryFile(delete=False)
	samfile = pysam.AlignmentFile(bam_file, "rb")
	for i in samfile:
		# Ignore unmapped reads
		if not i.is_unmapped:
			chrom = samfile.getrname(i.reference_id)
			if chrom in gdict.keys():
				# Determine which end of the read is the 5' end
				if i.is_reverse:
					strand = "-"
					startbp, endbp = i.reference_end - 3, i.reference_end + 3
				else:
					strand  = "+"
					startbp, endbp = i.reference_start - 3, i.reference_start + 3
				if startbp > 0 and endbp < gdict[chrom]:
					print >> outfile, "\t".join((str(i) for i in (chrom, startbp, endbp, 0, 0, strand)))
	outfile.close()
	return outfile.name

def genome_dic(g_file):
	"""
	Make a dictionary of chromosome sizes from a .chrom.sizes file
	e.g. {chrom_name:chrom_size}
	"""
	gdict = {}
	with open(g_file) as ifile:
		for i in ifile:
			i = i.split()
			gdict[i[0]] = int(i[1])
	return gdict

if __name__ == "__main__":
	
	parser = argparse.ArgumentParser(description='Calculates the 6-mer 5\' insertion bias for a NGS dataset')
	parser.add_argument("regions", help="BED file of the regions you want to exclude from calculating the bias. This is usually the DHSs.")
	parser.add_argument("reads", help="The sorted, indexed BAM file containing the DNase-seq data")
	parser.add_argument("genome_sequence", help="The sorted, indexed FASTA file containing the genome sequence")
	parser.add_argument("genomesize", help="The .chrom.sizes file containing chromosome sizes generated using something like  \"mysql --user=genome --host=genome-mysql.cse.ucsc.edu -A -e \"select chrom, size from hg19.chromInfo\"  > hg19.chrom.sizes\"")
	parser.add_argument("output", help="output file prefix to write the observed/expected ratios to (will append .txt and .pickle)")
	args  = parser.parse_args()
	
	test_bam        = args.reads
	test_bed        = args.regions
	genome_sequence = args.genome_sequence
	genome          = args.genome_size
	outfile         = args.output
	
	# First, pull all the 6mers surrounding 5' ends
	print("Determining transposition sites (roughly 60s per 1E6 reads)...")
	bed_file_for_6mers = generate_6mer_bed(test_bam, genome_dic(genome))
	all_6mers = pybedtools.BedTool(bed_file_for_6mers)
	
	# Can't guarantee peaks are sorted, so sort them
	peaks = pybedtools.BedTool(test_bed)	
	peaks = peaks.sort()
	
	# Get the transposition sites outside of the peaks only
	print("Filtering for those outside peaks")
	bg_6mers      = all_6mers.intersect(peaks,v=True)
	
	print("Generating shuffled background")
	shuf_bg_6mers = bg_6mers.shuffle(g=genome, noOverlapping=True, excl=peaks.fn)
		
	print("Generating FASTA file for 6mers...")
	genome_fasta  = pybedtools.BedTool(genome_sequence)
	observed_cuts = bg_6mers.sequence(fi=genome_fasta)
	
	print("Generating FASTA file for shuffled 6mers...")
	shuffled_cuts = shuf_bg_6mers.sequence(fi=genome_fasta)

	print("Getting 6mers for observed...")
	observed = get_kmers(observed_cuts.seqfn)

	print("Getting 6mers for shuffled...")
	expected = get_kmers(shuffled_cuts.seqfn)
	
	print("Calculating...")
	enriched = {i:observed[i]/float(expected[i]) for i in observed.keys()}

	print("Dumping bias txt file...")
	with open(outfile + ".txt", 'w') as ofile:
		for i in sorted(enriched.items(), key=operator.itemgetter(1)):
			print >> ofile, "\t".join(map(str,i))

	print("Writing bias pickle file...")
	totalsum = float(sum(enriched.values()))
	whatdic = {key:{'forward':val/totalsum,'reverse':enriched[rev_comp(key)]/totalsum} for key,val in enriched.iteritems()}
	with open(outfile + ".pickle", "w") as bias_file:
	    pickle.dump(whatdic,bias_file)
		
	os.remove(bed_file_for_6mers)



__author__ = 'Jason Piper'

import imp
current_version = imp.load_source('lol', 'pyDNase/_version.py').__version__

try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup
    from distutils.extension import Extension

setup(
    name='pyDNase',
    version=current_version,
    description='DNase-seq analysis library',
    long_description=open('README.rst',"rt").read(),
    author='Jason Piper',
    author_email='j.piper@warwick.ac.uk',
    url='http://jpiper.github.io/pyDNase',
    license='GPLv3',
    ext_modules = [Extension("pyDNase.footprinting.WellingtonC", sources = ["pyDNase/footprinting/WellingtonC.c"], extra_compile_args=['-std=c99'])],
    packages= [
        'pyDNase',
        'pyDNase.footprinting',
    ],

    install_requires=[
        # Not enforcing versions for numpy and matplotlib as they can be a bitch to upgrade
        "numpy", # Tested on >=1.5.0
        "matplotlib", # Tested on >=1.2
        "pysam >= 0.7.5",
        "clint >= 0.3.2",
    ],
    
    package_data = {'pyDNase':["data/*"]},
    
    scripts=[
        "pyDNase/scripts/dnase_cut_counter.py",
        "pyDNase/scripts/dnase_average_profile.py",
        "pyDNase/scripts/dnase_to_javatreeview.py",
        "pyDNase/scripts/dnase_wig_tracks.py",
        "pyDNase/scripts/wellington_footprints.py",
        "pyDNase/scripts/wellington_bootstrap.py",
        "pyDNase/scripts/dnase_to_JSON.py",
        "pyDNase/scripts/dnase_ddhs_scorer.py",
        "pyDNase/scripts/examples/example_footprint_scores.py",
        "pyDNase/scripts/dnase_to_JSON.py",
        "pyDNase/scripts/dnase_bias_estimator.py"],
    test_suite="test",
)


#!/usr/bin/env python

# Copyright (C) 2013 Jason Piper - j.piper@warwick.ac.uk
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import pyDNase
import numpy as np
import matplotlib as mpl
from clint.textui import progress, puts
#Required for headless operation
mpl.use('Agg')
import matplotlib.pyplot as plt
from matplotlib import rcParams

parser = argparse.ArgumentParser(description='Plots average profile of DNase (or Tn5 for ATAC-seq) activity surrounding a list of regions in a BED file')
parser.add_argument("-w", "--window_size", help="Size of flanking area around centre of the regions to plot (default: 100)",default=100,type=int)
parser.add_argument("-bf", "--bias-file", help="Location of the sorted, index",default = None,type=str)
parser.add_argument("-i",action="store_true", help="Ignores any strand information in BED file and plots data relative to reference strand",default=False)
parser.add_argument("-c",action="store_true", help="Combine the strand information into one graph",default=False)
parser.add_argument("-n",action="store_true", help="Normalise cut counts to a fraction peaks",default=False)
parser.add_argument("-b",action="store_true", help="Normalise for cutting bias",default=False)
parser.add_argument("-A",action="store_true", help="ATAC-seq mode",default=False)
parser.add_argument("regions", help="BED file of the regions you want to generate the average profile for")
parser.add_argument("reads", help="The BAM file containing the DNase-seq data")
parser.add_argument("output", help="filename to write the output to")
args  = parser.parse_args()

reads   = pyDNase.BAMHandler(args.reads,ATAC=args.A)
if args.b:
    if args.bias_file != None:
        freads   = pyDNase.BAMHandlerWithBias(pyDNase.FASTAHandler(args.bias_file),args.reads,ATAC=args.A)
    else:
        raise ValueError("No FASTA file provided for bias correction!")
regions = pyDNase.GenomicIntervalSet(args.regions)



#Set all strands to positive if "ignore strands" is enabled
if args.i:
    for each in regions:
        each.strand = "+"

puts("Resizing Regions to {0}".format(args.window_size))
regions.resizeRegions(args.window_size)

fw = []
rv = []
puts("Reading Data from BAM file...")
for each in progress.bar(regions):
    if sum(reads[each]["+"]) and sum(reads[each]["-"]):
        if args.b:
            try:
                fw.append(np.divide(reads[each]["+"],freads[each]["+"]))
                rv.append(np.divide(reads[each]["-"],freads[each]["-"]))
            except Exception:
                pass
        else:
            fw.append(reads[each]["+"])
            rv.append(reads[each]["-"])

if args.n:
    fw = [map(float,i)for i in fw]
    rv = [map(float,i) for i in rv]
    fw = [np.divide(np.subtract(i, min(i)), np.subtract(max(i) , min(i))) for i in fw]
    rv = [np.divide(np.subtract(i, min(i)), np.subtract(max(i) , min(i))) for i in rv]

if args.c:
    plt.plot(np.add(np.mean(fw,axis=0),np.mean(rv,axis=0)),c="red")
else:
    plt.plot(np.mean(fw,axis=0),c="red")
    plt.plot(np.mean(rv,axis=0),c="blue")

#Pad the axis out reads bit
rcParams['xtick.major.pad'] = 20 
rcParams['ytick.major.pad'] = 20

#Sort out the X axis ticks
ticks = [0,args.window_size,args.window_size*2]
labels = [-args.window_size,0,args.window_size]
plt.xticks(ticks, labels)

#Make the yaxis start from 0
plt.gca().set_ylim(0)

#Makes ticks only appear on the left hand side
plt.gca().yaxis.set_ticks_position('left')

#Remove top and right borders
plt.gca().spines['top'].set_visible(False)
plt.gca().spines['right'].set_visible(False)

plt.gca().tick_params(axis='both', which='major', labelsize=28, pad=12)

if args.bias_file:
    plt.gca().set_ylabel('Average DNase Activity\n (Observed/Expected)',size="32", multialignment='center')
else:
    if args.A:
        plt.gca().set_ylabel('Average Tn5 integrations',size="26", multialignment='center')
    else:
        plt.gca().set_ylabel('Average DNase activity',size="26", multialignment='center')
plt.savefig(args.output,bbox_inches='tight')



# Copyright (C) 2013 Jason Piper - j.piper@warwick.ac.uk
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import warnings, random
import numpy as np
import pyDNase
from . import WellingtonC

class wellington(object):

    def __init__(self,interval,reads,shoulder_sizes=range(35,36), footprint_sizes = range(11,26,2),FDR_cutoff=0.01,FDR_iterations=100,bonferroni=None):

        #Set up the model parameters
        self.shoulder_sizes  = shoulder_sizes
        self.footprint_sizes = footprint_sizes
        self.FDR_iterations  = FDR_iterations
        self.FDR_cutoff      = FDR_cutoff
        if bonferroni:
            self.bonferroni_factor = np.log(1.0/sum(reads.samfile.lengths))
        else:
            self.bonferroni_factor = None

        #Here we check if the interval is big enough for the parameters give
        if len(interval) < (max(shoulder_sizes)*2) + max(footprint_sizes):
            raise ValueError("The interval you're trying to footprint is smaller than the parameters you've passed.")

        self.interval        = interval
        if self.interval.strand is "-":
            warnings.warn("You're footprinting an interval on the reverse strand! "+
                          "You should be sure you know what you're doing as wellington was not designed for this! "+
                          "Ensure that all the intervals you provide to wellington are on the +ve strand for published behaviour",UserWarning)

        #We don't store the read object anymore as it isn't pickleable due it being a file handle.
        self.reads        = reads[self.interval]

        #"None" indicates that these haven't been calculated yet.
        self.storedscore   = None
        self.storedlengths = None
        self.FDR_Result    = None

    #This is a pretty hacky way to allow this object to be called from a multithreaded standpoint.
    #If you call the instanced object itself, it calculates the scores and the null scores, and returns
    #the entire object.
    def __call__(self):
        self.storedscore, self.storedlengths   = self.calculate()
        self.FDR_Result = self.FDRscore()
        return self

    #So I've made FDR_Value, lengths, and scores properties which are calculated ad hoc - they are no longer calculated
    #at instantiation
    @property
    def FDR_value(self):
        if self.FDR_Result != None:
            return self.FDR_Result
        else:
            self.FDR_Result = self.FDRscore()
            return self.FDR_Result
    @property
    def lengths(self):
        if self.storedlengths != None: return self.storedlengths
        else:
            self.storedscore, self.storedlengths   = self.calculate()
            return self.storedlengths

    @property
    def scores(self):
        if self.storedlengths != None: return self.storedscore
        else:
            self.storedscore, self.storedlengths   = self.calculate()
            return self.storedscore

    def FDRscore(self):
        return WellingtonC.percentile(sum([self.calculate(FDR=1)[0] for i in range(self.FDR_iterations)],[]),self.FDR_cutoff)


    def footprints(self, withCutoff=-30, merge = 1):
        """
        This returns reads GenomicIntervalSet with the intervals retrieved below the specific cutoff applied to the selected data
        """
        #This find the positions of all the ranges below the cutoff using reads new method
        ranges = []
        tempMLE, templogProb = np.array(self.lengths), np.array(self.scores)

        #Here we have some different logic for selecting the summits of footprints
        #TODO: Document this part

        while templogProb.min() < withCutoff:
            minimapos = templogProb.argmin()
            minimafplen = tempMLE[minimapos]
            minimaphalffplen = int(minimafplen)/2
            lbound = max(minimapos-(minimaphalffplen),0)
            rbound = min(minimapos+(minimaphalffplen),len(templogProb))
            ranges.append((lbound,rbound,templogProb.min(),minimafplen))
            templogProb[max(lbound-minimafplen,0):min(rbound+minimafplen,len(templogProb))] = 1


        returnSet = pyDNase.GenomicIntervalSet()
        #Merges overlapping ranges (TODO: documentation)
        if ranges:
            # This change here changes the way we merge footprints from the probability trace
            #TODO: Documentation
            if merge:
                merged_ranges = []
                while len(ranges):
                    #Find best score
                    sorted(ranges,key=lambda x:-x[2])
                    #Take the last value
                    best = ranges.pop()
                    merged_ranges.append(best)
                    #Check for overlapping regions and remove
                    new_ranges = []
                    for c, d, e, f in ranges:
                        if not c<=best[1]<=d:
                            new_ranges.append([c, d, e,f])
                    ranges = new_ranges
            else:
                merged_ranges = ranges
            #Creates reads GenomicIntervalSet and adds the footprints to them
            for i in merged_ranges:
                rstartbp = self.interval.startbp + i[0]
                #We must add one to the end base of the footprint to account for the BED file format
                rendbp   = self.interval.startbp + i[1] + 1
                region = pyDNase.GenomicInterval(self.interval.chromosome, rstartbp, rendbp, strand="+",score=i[2])
                returnSet += region
        return returnSet

    def calculate(self,FDR=0):
        #The passes the read data to the Cython wrapper for Wellington. This yields a roughly 4x speed improvement
        return WellingtonC.calculate(FDR,self.reads["+"],self.reads["-"],self.footprint_sizes,self.shoulder_sizes, self.bonferroni_factor)


class wellington1D(wellington):
    def calculate(self,FDR=0):
        return WellingtonC.calculate1D(FDR,self.reads["+"],self.reads["-"],self.footprint_sizes,self.shoulder_sizes, self.bonferroni_factor)
